#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <ctime>

#define elemsof(x) (sizeof(x)/sizeof(*(x)))

inline int imin(int x, int y) { return (x < y) ? x : y; }


// NOTE: Environment variable 'TLIMIT' holds wallclock time you may spend.


//=================================================================================================
// Datatypes:


// ATTRIBUTES/CHARACTERS

struct Attr_t {
    int     color   : 4;     // -2=undefined, -1=any, 0-7=red, green, blue etc.
    int     size    : 5;     // -2=undefined, 0-9=valid sizes.
    int     scores  : 3;     // 0-3 under scores
    int     bold    : 2;     // -1=either, 0=normal, 1=bold
    int     italic  : 2;     // -1=either, 0=normal, 1=italic
    int     emph    : 3;     // -1=either, 0-2=strength (where 2 is strong emphasis)
    int     tt      : 2;     // 0=normal, 1=tt
};


struct Attr : public Attr_t {
    Attr(void) {        // (default constructor defines initial value of parser)
        color = size = -2;
        scores = bold = italic = emph = tt = 0; }

    Attr(const Attr& attr) {
        *this = attr; }

    Attr(const Attr_t& attr) {
        color = attr.color; size = attr.size; scores = attr.scores;
        bold = attr.bold; italic = attr.italic; emph = attr.emph; tt = attr.tt; }

    bool operator == (const Attr& other) const {
        return other.color == color && other.size == size && other.scores == scores && other.bold == bold && other.italic == italic && other.emph == emph && other.tt == tt; }

    bool operator != (const Attr& other) const {
        return !(*this == other); }

    void reset(void) {  // The PL operation
        scores = bold = italic = emph = tt = 0; }

    int index(void) {   // Calclulate an index between 0-95. Color and size is not used!
        assert(bold != -1 && italic != -1 && emph != -1);
        return (unsigned)scores | ((unsigned)bold << 2) | ((unsigned)italic << 3)
               | ((unsigned)tt << 4) | ((unsigned)emph << 5); }

    Attr(int index) {
        color  = size = -2;
        scores = index        & 3;
        bold   = (index >> 2) & 1;
        italic = (index >> 3) & 1;
        tt     = (index >> 4) & 1;
        emph   = (index >> 5); }

    bool consistent(const Attr& other) const {
        return (color  == -1 || other.color  == -1 || color  == other.color )
            && (bold   == -1 || other.bold   == -1 || bold   == other.bold  )
            && (italic == -1 || other.italic == -1 || italic == other.italic)
            && (emph   == -1 || other.emph   == -1 || emph   == other.emph  )
            && size   == other.size
            && scores == other.scores
            && tt     == other.tt; }

    void incorporate(const Attr& other) {
        if (other.color  != -1) color  = other.color;
        if (other.bold   != -1) bold   = other.bold;
        if (other.italic != -1) italic = other.italic;
        if (other.emph   != -1) emph   = other.emph;
        size   = other.size;
        scores = other.scores;
        tt     = other.tt; }

    void fillup(void) {
        if (color  == -1) color  = -2;
        if (bold   == -1) bold   = 0;
        if (italic == -1) italic = 0;
        if (emph   == -1) emph   = 0; }

};


struct Char : public Attr {
    int     chr;

    Char(int chr_, const Attr& attr) : Attr(attr), chr(chr_) { }
};


// SYNTAX TREE

enum TreeType { syn_Juxt, syn_Attr, syn_Text, syn_XJuxt };

typedef struct Tree_t* Tree;
struct Tree_t {
    TreeType type;

#ifdef OLD_GCC      // buggy...
    Tree    left;
    Tree    right;
    Attr_t  attr;
    Tree    tree;
    char*   text;
#else
    union {
        struct {
            Tree    left;
            Tree    right;
        };
        struct {
            Attr_t  attr;
            Tree    tree;
        };
        char*   text;
    };
#endif
};


vector <Tree>   all_trees;

void disposeTrees(void) {
    for (unsigned i = 0; i < all_trees.size(); i++){
        if (all_trees[i]->type == syn_Text) free(all_trees[i]->text);
        delete all_trees[i]; }
    all_trees.resize(0); }

Tree Tree_new(Tree left, Tree right) {
    Tree    tmp = new Tree_t; all_trees.push_back(tmp);
    tmp->type   = syn_Juxt;
    tmp->left   = left;
    tmp->right  = right;
    return tmp; }

Tree Tree_newX(Tree left, Tree right) {     // (don't ask...)
    Tree    tmp = new Tree_t; all_trees.push_back(tmp);
    tmp->type   = syn_XJuxt;
    tmp->left   = left;
    tmp->right  = right;
    return tmp; }

Tree Tree_new(Attr attr, Tree tree) {
    Tree    tmp = new Tree_t; all_trees.push_back(tmp);
    tmp->type   = syn_Attr;
    tmp->attr   = attr;
    tmp->tree   = tree;
    return tmp; }

Tree Tree_new(char* text) {
    Tree    tmp = new Tree_t; all_trees.push_back(tmp);
    tmp->type   = syn_Text;
    tmp->text   = text;
    return tmp; }


// WORDS

struct Word {       // (Has been generalized at bit)

    Attr    attr;   // Attribute (that should be) provided by context.
    Tree    tree;

    Word(void) { }

    Word(const Attr& attr_, char* text_) {
        attr = attr_;
        tree = Tree_new(strdup(text_));
        assert(attr.color != -1 && attr.bold != -1 && attr.italic != -1  && attr.emph != -1); }

    Word(const Attr& attr_, Tree tree_) {
        attr = attr_;
        tree = tree_; }
};


// TAGS

enum Tag {
    tag_B , tag_EM, tag_I , tag_PL, tag_S , tag_TT, tag_U , tag_0 , tag_1 , tag_2 , tag_3 , tag_4 , tag_5 ,
    tag_6 , tag_7 , tag_8 , tag_9 , tag_r , tag_g , tag_b , tag_c , tag_m , tag_y , tag_k , tag_w };

const char* on_tag [] = {
    "<B>" , "<EM>", "<I>" , "<PL>", "<S>" , "<TT>", "<U>" , "<0>" , "<1>" , "<2>" , "<3>" , "<4>" , "<5>" ,
    "<6>" , "<7>" , "<8>" , "<9>" , "<r>" , "<g>" , "<b>" , "<c>" , "<m>" , "<y>" , "<k>" , "<w>" };

const char* off_tag [] = {
    "</B>" , "</EM>", "</I>" , "</PL>", "</S>" , "</TT>", "</U>" , "</0>" , "</1>" , "</2>" , "</3>" , "</4>" , "</5>" ,
    "</6>" , "</7>" , "</8>" , "</9>" , "</r>" , "</g>" , "</b>" , "</c>" , "</m>" , "</y>" , "</k>" , "</w>" };


// TOKENS

enum TokenType { tok_ON, tok_OFF, tok_CHAR };
struct Token {
    TokenType   type : 2;
    Tag         tag  : 6;
    char        chr  : 8;
};
inline Token Token_on  (Tag tag) { Token t; t.type = tok_ON;   t.tag = tag; return t; };
inline Token Token_off (Tag tag) { Token t; t.type = tok_OFF;  t.tag = tag; return t; };
inline Token Token_char(int chr) { Token t; t.type = tok_CHAR; t.chr = chr; return t; };


//=================================================================================================
// Helpers:


int color_letter[] = { 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w' };


char is_ws[256] = { 0,0,0,0,0,0,0,0,0,1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1 };
#define isWS(chr) ((bool)is_ws[chr])


//=================================================================================================
// Debug:


void dump(Attr attr)
{
//  printf("color: %c   size: %c   scores: %c   bold: %c   italic: %c   emph: %c   tt: %c",
    printf("%c%c%c%c%c%c%c",
           (attr.color == -2) ? '@' : (attr.color == -1) ? '*' : color_letter[attr.color],
           (attr.size  == -2) ? '@' : '0' + attr.size,
           '0' + attr.scores,
           (attr.bold   == 0) ? 'b' : (attr.bold   == 1) ? 'B' : '*',
           (attr.italic == 0) ? 'i' : (attr.italic == 1) ? 'I' : '*',
           (attr.emph   == 0) ? 'e' : (attr.emph   == 1) ? 'E' : (attr.emph == 2) ? '!' : '*',
           (attr.tt     == 0) ? 't' : (attr.tt     == 1) ? 'T' : '*');
}


void dump(Char c)
{
    int     chr = c.chr;
    if      (chr == '\t') chr = '';
    else if (chr == '\r') chr = '';
    else if (chr == '\n') chr = '';
    else if (chr == ' ' ) chr = '';

    printf("%c : ", chr);
    dump((Attr)c);
    printf("\n");
}


void dumpCompact(Attr attr)
{
    if ( attr.color < 0 && attr.size < 0 && attr.scores == 0
      && attr.bold != 1 && attr.italic != 1 && attr.emph != 1 && attr.emph != 2 && attr.tt != 1)
        printf("-");
    else{
        if (attr.color  >= 0) printf("%c", color_letter[attr.color]);
        if (attr.size   >= 0) printf("%d", attr.size);
        for (int i = 0; i < attr.scores; i++) printf("_");
        if (attr.bold   == 1) printf("B");
        if (attr.italic == 1) printf("I");
        if (attr.emph   == 1) printf("E"); else if (attr.emph == 2) printf("S");
        if (attr.tt     == 1) printf("T");
    }
}


void dump(Tree tree)
{
    switch (tree->type){
    case syn_Juxt:
        dump(tree->left);
        printf(" ++ ");
        dump(tree->right);
        break;
    case syn_Attr:
        dumpCompact(tree->attr);
        printf("(");
        dump(tree->tree);
        printf(")");
        break;
    case syn_Text:
        printf("\"%s\"", tree->text);
        break;
    default: assert(0); }
}


void dump(const vector<Word>& words)
{
    for (int i = 0; i < (int)words.size(); i++){
        dumpCompact(words[i].attr);
        printf(" : ");
        dump(words[i].tree);
        printf("\n");
    }
}


//=================================================================================================
// Parsing:


void parse(char* input, vector<Char>& doc)
{
    vector<Attr>    states;
    Attr            state;

    while (*input != '\0') {
        switch (*input){
        case '<' :
            #define PARSE(nchars, action) input += nchars; states.push_back(state); action; break;

            input++;
            switch(*input){
            case 'T': PARSE(3, state.tt     = 1)
            case 'B': PARSE(2, state.bold   = 1)
            case 'I': PARSE(2, state.italic = 1)
            case 'S': PARSE(2, state.emph   = 2)
            case 'E': PARSE(3, if (state.emph < 2) state.emph = 1 - state.emph)
            case 'U': PARSE(2, if (state.scores < 3) state.scores++)
            case 'P': PARSE(3, state.reset())

            case '0': PARSE(2, state.size = 0)
            case '1': PARSE(2, state.size = 1)
            case '2': PARSE(2, state.size = 2)
            case '3': PARSE(2, state.size = 3)
            case '4': PARSE(2, state.size = 4)
            case '5': PARSE(2, state.size = 5)
            case '6': PARSE(2, state.size = 6)
            case '7': PARSE(2, state.size = 7)
            case '8': PARSE(2, state.size = 8)
            case '9': PARSE(2, state.size = 9)

            case 'r': PARSE(2, state.color = 0)
            case 'g': PARSE(2, state.color = 1)
            case 'b': PARSE(2, state.color = 2)
            case 'c': PARSE(2, state.color = 3)
            case 'm': PARSE(2, state.color = 4)
            case 'y': PARSE(2, state.color = 5)
            case 'k': PARSE(2, state.color = 6)
            case 'w': PARSE(2, state.color = 7)

            case '/':
                state = states[states.size()-1];
                states.pop_back();
                do input++; while (*input != '>');
                input++;
                break;

            default: /**/fprintf(stderr, "CHAR: %d %c\n", *input, *input); assert(false); }

            break;
            #undef PARSE

        case '>' :  assert(false);

        case ' ' :
        case '\t':
        case '\r':
        case '\n': {
            Char    c(*input, state);
            c.bold = c.italic = c.emph = -1;
            if (state.scores == 0)
                c.color = -1;
            if (state.tt == 0)
                c.chr = ' ';        // <= Replace all WS by SPC (if in non-TT mode)
            doc.push_back(c);
            input++;
            } break;

        default: {
            Char    c(*input, state);
            doc.push_back(c);
            input++;
            }
        }
    }
}


void compressWS(const vector<Char>& in, vector<Char>& out)
{
    if (in.size() == 0) return;

    out.push_back(in[0]);
    for (unsigned i = 1; i < in.size(); i++){
        if (isWS(in[i].chr)){
            if (in[i].tt == 1)
                out.push_back(in[i]);
            else if (out.back().chr == ' ' && (Attr)in[i] == (Attr)out.back())
                /*nothing*/;
            else
                out.push_back(Char(' ', Attr(in[i])));
        }else
            out.push_back(in[i]);
    }
}


char* readStdin(void)
{
    char*   buf      = (char*)malloc(1);
    int     capacity = 1;
    int     size     = 0;
    int     chr;
    while (chr = getchar(),
           chr != EOF){
        if (size >= capacity){ capacity *= 2; buf = (char*)realloc(buf, capacity); }
        buf[size++] = chr;
    }
    if (size >= capacity){ capacity *= 2; buf = (char*)realloc(buf, capacity); }
    buf[size++] = '\0';
    return buf;
}


//=================================================================================================
// Costs:


#define LARGE_COST 10000    // (any number bigger than the longest possible command sequence will do)

int cost_table[96][96];


inline bool haveToReset(Attr from, Attr to) {
    return ( (from.tt     == 1 && to.tt         == 0)
          || (from.italic == 1 && to.italic     == 0)
          || (from.bold   == 1 && to.bold       == 0)
          || (from.emph   == 2 && (to.emph & 2) == 0)
          || (from.scores > to.scores)); }


int cost_helper(Attr from, Attr to)
{
    assert(from.italic != -1 && from.bold != -1 && from.emph != -1);

    int     cost;

    if (haveToReset(from, to))
        from.reset(),
        cost = 9 /*PL_cost*/;
    else
        cost = 0;

    if (to.color == -2 && from.color != -2) return LARGE_COST;
    if (to.size  == -2 && from.size  != -2) return LARGE_COST;

    if (to.color  != -1 && from.color  != to.color ) cost += 7 /*color_cost*/;
    if (                   from.size   != to.size  ) cost += 7 /*size_cost*/;
    if (                   from.scores != to.scores) cost += 7 /*U_cost*/ * (to.scores - from.scores);
    if (to.bold   != -1 && from.bold   != to.bold  ) cost += 7 /*B_cost*/;
    if (to.italic != -1 && from.italic != to.italic) cost += 7 /*I_cost*/;
    if (to.emph   != -1 && from.emph   != to.emph  ) cost += (to.emph == 2) ? 7 /*S_cost*/ : 9 /*EM_cost*/;
    if (                   from.tt     != to.tt    ) cost += 9 /*TT_cost*/;

    return cost;
}


void cost_init(void)
{
    int     i, j;

    for (i = 0; i < 96; i++){
        for (j = 0; j < 96; j++){
            Attr    from(i);
            Attr    to  (j);
            cost_table[i][j] = cost_helper(from, to);
        }
    }
}


inline int cost(Attr from, Attr to)
{
    if (to.color == -2 && from.color != -2) return LARGE_COST;
    if (to.size  == -2 && from.size  != -2) return LARGE_COST;
    return cost_table[from.index()][to.index()]
         + ((to.color != from.color) ? 7 /*color_cost*/ : 0)
         + ((to.size  != from.size ) ? 7 /*size_cost */ : 0);
}


//=================================================================================================
// Optimizer:


void chunkDoc(const vector<Char>& doc, vector<Word>& words)
{
    Attr            state;
    int             i;
    vector<char>    text;

    for (i = 0; i < (int)doc.size(); i++){
        if (state.consistent((Attr)doc[i])){
            text.push_back(doc[i].chr);
            state.incorporate((Attr)doc[i]);
        }else{
            text.push_back('\0');
            state.fillup();
            words.push_back(Word(state, &text[0]));
            text.clear();
            text.push_back(doc[i].chr);
            state = (Attr)doc[i];
        }
    }
    text.push_back('\0');
    state.fillup();
    words.push_back(Word(state, &text[0]));
}



// Behind must have room for ahead_sz more elements. These element may be ahead itself! (I think...)
//
int lookahead(Attr* behind, int behind_sz, Attr* ahead, int ahead_sz, int* ret_answer = NULL)
{
    int     delta, delta_best;
    int     total, total_best;
    int     i    , i_best;
    Attr    tmp;

    if (ahead_sz == 0){
        assert(ret_answer == NULL);
        return 0; }

    delta_best = cost(behind[behind_sz-1], ahead[0]);   // If LARGE_NUMBER, could handle the situation a bit more efficient than implemented here...

    tmp = behind[behind_sz];
    behind[behind_sz] = ahead[0];
    total_best = delta_best + lookahead(behind, behind_sz+1, ahead+1, ahead_sz-1);
    behind[behind_sz] = tmp;
    i_best     = behind_sz-1;

    for (i = behind_sz-2; i >= 0; i--){
        delta = cost(behind[i], ahead[0]);
        if (delta < delta_best){
            delta_best = delta;

            tmp = behind[i+1];
            behind[i+1] = ahead[0];
            total = delta + lookahead(behind, i+2, ahead+1, ahead_sz-1);
            behind[i+1] = tmp;

            if (total < total_best)
                total_best = total,
                i_best     = i;
        }
    }

    if (ret_answer != NULL){
        *ret_answer = (behind_sz-1) - i_best;
    }

    return total_best;
}


int lookahead_wrapper(vector<Word>& result, vector<Word>& words, int index, int history_size, int lookahead_size)
{
    if (history_size   > (int)result.size()     ) history_size   = result.size();
    if (lookahead_size > (int)words.size()-index) lookahead_size = words.size()-index;

    Attr    buf[history_size + lookahead_size];
    int     i, j, pops;

    for (i = history_size-1, j = result.size()-1; i >= 0; i--, j--)
        buf[i] = result[j].attr;
    for (i = history_size, j = index; i < history_size + lookahead_size; i++, j++)
        buf[i] = words[j].attr;

    lookahead(buf, history_size, buf + history_size, lookahead_size, &pops);
    return pops;
}


void stackPerform(Attr from, Attr to, vector<Word>& stack)
{
    assert(from.italic != -1 && from.bold != -1 && from.emph != -1);

    if (haveToReset(from, to))
        from.reset(),
        stack.push_back(Word(from, ""));

    if (to.color != -1 && from.color != to.color){
        assert(to.color != -2);
        from.color = to.color;
        stack.push_back(Word(from, "")); }

    if (from.size != to.size){
        assert(to.size != -2);
        from.size = to.size;
        stack.push_back(Word(from, "")); }

    if (from.scores != to.scores)
        for (int i = (to.scores - from.scores); i != 0; i--)
            from.scores++,
            stack.push_back(Word(from, ""));

    if (to.bold != -1 && from.bold != to.bold)
        assert(to.bold == 1),
        from.bold = 1,
        stack.push_back(Word(from, ""));

    if (to.italic != -1 && from.italic != to.italic)
        assert(to.italic == 1),
        from.italic = 1,
        stack.push_back(Word(from, ""));

    if (to.emph != -1 && from.emph != to.emph){
        from.emph = to.emph;
        stack.push_back(Word(from, "")); }

    if (from.tt != to.tt)
        assert(to.tt == 1),
        from.tt = 1,
        stack.push_back(Word(from, ""));

}


Tree optimize(vector<Word>& words, int history_size, int lookahead_size)
{
    Word    tmp;
    #define POP \
          ( tmp = result.back(), \
            result.pop_back(), \
            result.back() = Word(result.back().attr, Tree_new(result.back().tree, Tree_new(tmp.attr, tmp.tree))) \
          )

    assert(words.size() > 0);
    vector<Word>    result;
    int             i, pops;

    result.push_back(Word(Attr(), ""));

    for (i = 0; i < (int)words.size(); i++){
        // Forced pops:
        while ( (words[i].attr.color == -2 && result.back().attr.color != -2)
             || (words[i].attr.size  == -2 && result.back().attr.size  != -2))
            POP;

        // Voluntary pops:
        pops = lookahead_wrapper(result, words, i, history_size, lookahead_size);

        while (pops > 0)
            POP, pops--;

        if (result.back().attr == words[i].attr)    // (two words with same attribute next to each other => merge)
            result.back() = Word(result.back().attr, Tree_new(result.back().tree, words[i].tree));
        else{                                       // (just push the new word)
            result.push_back(words[i]);
        }
    }

    while (result.size() > 1)
        POP;

    return result[0].tree;
}


//=================================================================================================
// Realizer:


void realize(Tree tree, vector<char>& out, Attr state = Attr());



void perform(Attr from, Attr to, vector<Tag>& tags)
{
    assert(from.italic != -1 && from.bold != -1 && from.emph != -1);

    if (haveToReset(from, to))
        tags.push_back(tag_PL),
        from.reset();

    if (to.color != -1 && from.color != to.color){
        assert(to.color != -2);
        tags.push_back((Tag)((int)tag_r + to.color));
        from.color = to.color; }

    if (from.size != to.size){
        assert(to.size != -2);
        tags.push_back((Tag)((int)tag_0 + to.size));
        from.size = to.size; }

    if (from.scores != to.scores)
        for (int i = (to.scores - from.scores); i != 0; i--)
            tags.push_back(tag_U),
            from.scores++;

    if (to.bold != -1 && from.bold != to.bold)
        assert(to.bold == 1),
        tags.push_back(tag_B),
        from.bold = 1;

    if (to.italic != -1 && from.italic != to.italic)
        assert(to.italic == 1),
        tags.push_back(tag_I),
        from.italic = 1;

    if (to.emph != -1 && from.emph != to.emph){
        if (to.emph == 2)
            tags.push_back(tag_S);
        else
            assert(from.emph != 2),
            tags.push_back(tag_EM);
        from.emph = to.emph; }

    if (from.tt != to.tt)
        assert(to.tt == 1),
        tags.push_back(tag_TT),
        from.tt = 1;

}


inline void push_string(char* text, vector<char>& out) {
    for (int i = 0; text[i] != '\0'; i++)
        out.push_back(text[i]); }


inline void push_onTag(Tag tag, vector<char>& out) {
    for (int i = 0; on_tag[tag][i] != '\0'; i++)
        out.push_back(on_tag[tag][i]); }


inline void push_offTag(Tag tag, vector<char>& out) {
    for (int i = 0; off_tag[tag][i] != '\0'; i++)
        out.push_back(off_tag[tag][i]); }


void push_onTags(const vector<Tag>& tags, vector<char>& out) {
    for (int i = 0; i < (int)tags.size(); i++)
        push_onTag(tags[i], out); }


void push_offTags(const vector<Tag>& tags, vector<char>& out) {
    for (int i = tags.size()-1; i >= 0; i--)
        push_offTag(tags[i], out); }



// (returns a list of attributed trees)
void collectAttrList(Tree tree, Attr state, vector<Tree>& attr_list)
{
    switch (tree->type){
    case syn_Juxt:
        collectAttrList(tree->left , state, attr_list);
        collectAttrList(tree->right, state, attr_list);
        break;
    case syn_Attr: {
        attr_list.push_back(tree);
        break; }
    case syn_Text:
        attr_list.push_back(Tree_new(state, tree));
        break;
    default: assert(0); }
}


inline Attr inCommon(Attr state, Attr l, Attr r)
{
    assert(!haveToReset(state, l) && !haveToReset(state, r));
    if (l.color == r.color) state.color = l.color;
    if (l.size  == r.size ) state.size  = l.size ;
    if (l.emph  == r.emph)  state.emph  = l.emph ;
    state.scores = imin(l.scores, r.scores);
    state.bold   = imin(l.bold  , r.bold  );
    state.italic = imin(l.italic, r.italic);
    state.tt     = imin(l.tt    , r.tt    );
    return state;
}


void blend_helper(Attr state, vector<Tree>& list, int first, int last, bool reset, vector<char>& out)
{
    if (first == last){
        realize(list[first], out, state);
        return; }

    if (reset)
        push_onTag(tag_PL, out),
        state.reset();

    vector<Tree>    stack;
#if 1
    stack.push_back(list[first]);
    for (int i = first+1; i <= last; i++){
        if (!(inCommon(state, stack.back()->attr, list[i]->attr) == state))
            stack.back() = Tree_new(inCommon(state, stack.back()->attr, list[i]->attr), Tree_newX(stack.back(), list[i]));
        else
            stack.push_back(list[i]);
    }
    for (int i = 0; i < (int)stack.size(); i++)
        realize(stack[i], out, state);
#else
    stack.push_back(list[last]);
    for (int i = last-1; i >= first; i--){
        if (!(inCommon(state, stack.back()->attr, list[i]->attr) == state))
            stack.back() = Tree_new(inCommon(state, stack.back()->attr, list[i]->attr), Tree_newX(list[i], stack.back()));
        else
            stack.push_back(list[i]);
    }
    for (int i = stack.size()-1; i >= 0; i--)
        realize(stack[i], out, state);
#endif

    if (reset)
        push_offTag(tag_PL, out);
}


void blend(Attr init_state, vector<Tree>& list, vector<char>& out)
{
    bool    reset;
    int     i, j;

    assert(list.size() >= 2);

    reset = haveToReset(init_state, list[0]->attr);
    i = 0;
    for (j = 1; j < (int)list.size(); j++){
        if (haveToReset(init_state, list[j]->attr) != reset){
            blend_helper(init_state, list, i, j-1, reset, out);
            i = j;
            reset = !reset;
        }
    }
    blend_helper(init_state, list, i, j-1, reset, out);
}


void realize(Tree tree, vector<char>& out, Attr state)
{
    switch (tree->type){
    case syn_Juxt: {
        vector<Tree>    list;
        collectAttrList(tree, state, list);
        blend(state, list, out);
        } break;
    case syn_XJuxt:
        realize(tree->left , out, state);
        realize(tree->right, out, state);
        break;
    case syn_Attr: {
        vector<Tag> tags;
        perform(state, tree->attr, tags);
        push_onTags(tags, out);
        realize(tree->tree, out, tree->attr);
        push_offTags(tags, out);
        } break;
    case syn_Text:
        push_string(tree->text, out);
        break;
    default: assert(0); }
}


//=================================================================================================
// Old optimizer: (sometimes better)


#define MAX_POPS 10

void perform(Attr& state, Attr to, vector<Tag>& tag_stack, vector<Attr>& state_stack, vector<Token>& out)
{
    #define PUSH(tag) (tag_stack.push_back(tag), state_stack.push_back(state), out.push_back(Token_on(tag)))

    assert(state.italic != -1);
    assert(state.bold != -1);
    assert(state.emph != -1);

    if (haveToReset(state, to))
        PUSH(tag_PL),
        state.reset();

    if (to.color != -1 && state.color != to.color){
        assert(to.color != -2);
        PUSH((Tag)((int)tag_r + to.color));
        state.color = to.color; }

    if (state.size != to.size){
        assert(to.size != -2);
        PUSH((Tag)((int)tag_0 + to.size));
        state.size = to.size; }

    if (state.scores != to.scores)
        for (int i = (to.scores - state.scores); i != 0; i--)
            PUSH(tag_U),
            state.scores++;

    if (to.bold != -1 && state.bold != to.bold)
        assert(to.bold == 1),
        PUSH(tag_B),
        state.bold = 1;

    if (to.italic != -1 && state.italic != to.italic)
        assert(to.italic == 1),
        PUSH(tag_I),
        state.italic = 1;

    if (to.emph != -1 && state.emph != to.emph){
        if (to.emph == 2)
            PUSH(tag_S);
        else
            assert(state.emph != 2),
            PUSH(tag_EM);
        state.emph = to.emph; }

    if (state.tt != to.tt)
        assert(to.tt == 1),
        PUSH(tag_TT),
        state.tt = 1;

    #undef PUSH
}


void perform_forcedPops(Attr& state, Attr to, vector<Tag>& tag_stack, vector<Attr>& state_stack, vector<Token>& out)
{
    if (to.color == -2){
        while (state.color != -2){
            out.push_back(Token_off(tag_stack.back()));
            tag_stack.pop_back();
            assert(state_stack.size() > 0);
            state = state_stack.back();
            state_stack.pop_back(); }
    }

    if (to.size == -2){
        while (state.size != -2){
            out.push_back(Token_off(tag_stack.back()));
            tag_stack.pop_back();
            assert(state_stack.size() > 0);
            state = state_stack.back();
            state_stack.pop_back(); }
    }
}


void maybePop(Attr& state, Attr to, vector<Tag>& tag_stack, vector<Attr>& state_stack, vector<Token>& out)
{
    if (state == to || state_stack.size() == 0) return;

    int     best;
    int     best_n;
    int     i, n;
    int     d;

    best = cost_helper(state, to);
    best_n = 0;
    for (i = (int)state_stack.size()-1, n = 1; i >= 0 && n <= MAX_POPS; i--, n++){
        d = cost_helper(state_stack[i], to);
        if (d < best)
            best = d, best_n = n;
    }

    for (i = 0; i < best_n; i++){
        out.push_back(Token_off(tag_stack.back()));
        tag_stack.pop_back();
        assert(state_stack.size() > 0);
        state = state_stack.back();
        state_stack.pop_back();
    }
}


void greedy(const vector<Char>& doc, vector<Token>& out)
{
    Attr            state;
    vector<Tag>     tags;
    vector<Attr>    states;
    int             i;

    for (i = 0; i < (int)doc.size(); i++){
        perform_forcedPops(state, doc[i], tags, states, out);
        maybePop          (state, doc[i], tags, states, out);
        perform           (state, doc[i], tags, states, out);
        out.push_back(Token_char(doc[i].chr));
    }

    for (int j = (int)tags.size()-1; j >= 0; j--)
        out.push_back(Token_off(tags[j]));
}


void print(const vector<Token>& toks, vector<char>& out) {
    for (int i = 0; i < (int)toks.size(); i++){
        switch (toks[i].type){
        case tok_ON  : push_onTag (toks[i].tag, out)   ; break;
        case tok_OFF : push_offTag(toks[i].tag, out)   ; break;
        case tok_CHAR: out.push_back((char)toks[i].chr); break;
        default: assert(false); }
    }
}


//=================================================================================================
// More debug:


void ReMark(vector<Char> doc, vector<char>& out)
{
    vector<Token>   toks;
    greedy(doc, toks);
    print(toks, out);
}


void ReMarkToo(vector<Char> doc, int history, int lookahead, vector<char>& out)
{
    Tree            result;
    vector<Word>    words;

    chunkDoc(doc, words);
    result = optimize(words, history, lookahead);
    realize(result, out);
    disposeTrees();
}

int main(int argc, char** argv)
{
    int     history   = 10;
    int     lookahead = 5;
    int     timeout   = 100000000;

    if (argc >= 2) history   = atoi(argv[1]);
    if (argc >= 3) lookahead = atoi(argv[2]);
    if (argc >= 4) timeout   = atoi(argv[3]);   // <== Not implemented

    cost_init();
    srand(time(0));

    vector<char>    out1, out2,* best;
    vector<Char>    doc, doc2;

    char*   orig = readStdin();
    parse(orig, doc);

    compressWS(doc, doc2);

    // Begin...

    ReMark(doc2, out1);
    //fprintf(stderr, "Greedy: %d\n", out1.size());

    if (doc2.size() < 20000){
        ReMarkToo(doc2, 10, 5, out2);
        //fprintf(stderr, "New   : %d\n", out2.size());
        best = (out1.size() < out2.size()) ? &out1 : &out2;
    }else
        best = &out1;

    // Done!

    if (strlen(orig) < best->size())   // Sanity check
        printf("%s", orig);
    else{
        best->push_back('\0');
        printf("%s", &(*best)[0]);
    }
    free(orig);

    return 0;
}
